/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2014, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComToneMapping.h
    \brief    Tone Mapping class (header)
*/

#ifndef __TCOMCOLORMAPPINGALGO__
#define __TCOMCOLORMAPPINGALGO__


// Include files
#include "TComPicYuv.h"
#include "TComLut3D.h"

typedef struct {
  Double  dY, dU, dV;
} nodeD; // node

typedef struct {
  Double  dtchCoeffMatrix[4][7];
  Int     nbPts;

  //Double  GainOffsetMatrix[3][4];
  Double  **GainOffsetMatrix;
  Bool    bOk;

  nodeD   nD[2][2][2];

  Double  sumDiffY, sumDiffU, sumDiffV;
#if PIECEWISE_LINEAR_1DLUT
  Double  **invGainOffsetMatrix;
#endif
} ColorData;

// ====================================================================================================================
// Class definition
// ====================================================================================================================
/// tone mapping class
class TComColorMappingAlgo
{
  Bool        m_bVerbose;

  Int         m_nbp;
  Int         m_shift_div;
  ColorData*  m_colorData;

  //Void        compute_GO_model( ColorData* pColorData, Double GainOffsetMatrix[3][4] );
  Void        compute_GO_model( ColorData* pColorData, Double **GainOffsetMatrix );

public:
  TComColorMappingAlgo( Int iNbpCode );
  ~TComColorMappingAlgo();

  Void  razColorData();
  Void  appendColorSamples( TComPicYuv* pcBasePic0,  TComPicYuv* pcBasePic1, Int stepX, Int stepY, Int border, Int InputBitDepth=8, Bool b2x=false, Int appendBitDepth=8 );
  Void  compute_GO_model( Int InputBitDepth );
  //Void  modelMatrix( Double GainOffsetMatrix[3][4], nodeD *nodeIn, nodeD *nodeOut, Int InputBitDepth );
  Void  modelMatrix( Double **GainOffsetMatrix, nodeD *nodeIn, nodeD *nodeOut, Int InputBitDepth );

  Void	  appendColorSamplesDiff( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePicDiff, Int stepX, Int stepY, Int InputBitDepth=8 );
  Void    orderingDiff( Int, Bool bVerbose, Int *piOrderY, Int *piOrderU, Int *piOrderV, UInt64 *piSDY=NULL, UInt64 *piSDU=NULL, UInt64 *piSDV=NULL );

  inline  ColorData*  getColorData( int r, int g, int b );
  inline  ColorData*  getColorData( int n );
  inline  Void        getColorDataCoordinates( int n, int *r, int *g, int *b );

  Int                 getNbp()  { return m_nbp;};

  Void  setVerbose(Bool b)  { m_bVerbose = b; };
};

inline  ColorData*  TComColorMappingAlgo::getColorData( int r, int g, int b )
{
  return( m_colorData + r*m_nbp*m_nbp + g*m_nbp + b );
}

inline  ColorData*  TComColorMappingAlgo::getColorData( int n )
{
  return( m_colorData + n );
}

inline  Void  TComColorMappingAlgo::getColorDataCoordinates( int n, int *r, int *g, int *b )
{
  (*r)  = n / (m_nbp*m_nbp);
  (*g)  = (n - (*r)*(m_nbp*m_nbp)) / m_nbp;
  (*b)  = n - (*r)*(m_nbp*m_nbp) - (*g)*m_nbp;
}



class TComColorMappingMainAlgo
{
  Bool        m_verbose;
  Int         m_nbp;
  Int         m_nbpCode;
  Int         m_inputBitDepth;  // BL bit-depth
  Int         m_outputBitDepth; // EL bit-depth
  Int         m_appendBitDepth; // bit-depth of data used for computing LUT
  Int         m_lutBitDepth;
  Int         m_subSampling;
  Int         m_border;
  Int         m_nbPtsMinThreshold;
  TComLut3D*  m_lutColor;

  TComColorMappingAlgo**  pTabColorMapping;
  TComLut3D**         pTabLutColor;

  Void  compute_GO_model( TComColorMappingAlgo* pColorMapping, TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1, Int nbpCode );
  Void  compute_3D_Lut( TComLut3D* lutColor, TComColorMappingAlgo* pColorMapping, Int nbp, Int denomShift );
  Void  getRecursiveNode( node* nodeIn, TComColorMappingAlgo** pTabColorMapping, Int nbpCode, node* nodeOut, Int denomShift, Int inputBitDepth );

  Void  downScaleComponent2x2( const Pel* pSrc, Pel* pDest, const Int iSrcStride, const Int iDestStride, const Int iSrcWidth, const Int iSrcHeight, const Int ibitDepth, const Int obitDepth );

  inline Short  xClip( Short x , Int bitdepth );

public:
  TComColorMappingMainAlgo( Int iNbpCode, Int iInputBitDepth, Int iAppendBitDepth, Int iLutBitDepth, Int iSubSampling, Int iBorder, Bool bVerbose=false );
  ~TComColorMappingMainAlgo();

  Void  downScalePic( TComPicYuv* pcYuvSrc, TComPicYuv* pcYuvDest, Int ibitDepth, Int obitDepth );

  Void  computeCLUT( TComPicYuv* pcBasePic0, TComPicYuv* pcBasePic1 );

  TComLut3D*  getLut()  { return m_lutColor; };
};

inline Short TComColorMappingMainAlgo::xClip( Short x , Int bitdepth )
{ 
  Int max = (Int)(1 << bitdepth) - 1;

  Short pel = (Pel)( (x < 0) ? 0 : ((x > max) ? max : x) );
  return( pel );
}

#endif 
