/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

// Include files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TComToneMapping.h"
#include "TComLut3D.h"
//#include "TComLut3DDouble.h"

static  UChar nbpCode[7]={ 0, 2, 3, 5, 9, 17, 33 };

static  Void  assert( Char *comment, Bool bTest )
{
  if ( !bTest ) {
    printf("assertion failed:\t%s\n", comment);
    exit(0);
  }
}

// ====================================================================================================================
// Class definition
// ====================================================================================================================
TComToneMapping::TComToneMapping( Int iModel, Int iNbpCode, Char *cToneMappingFile )
{
  m_bVerbose        = false;

  m_toneMappingFile = cToneMappingFile;
  m_nbpCode         = iNbpCode;

  m_gain          = NULL;
  m_offset        = NULL;
  m_lutRef        = NULL;
  m_lutCoded      = NULL;
  m_lutResiduals  = NULL;

  m_denom_shift   = 0;

  if ( cToneMappingFile ) loadFile( cToneMappingFile );
  else                    m_model = iModel;

	m_pcTempPic	= NULL;
  m_errQ      = 1;

  switch( m_model ) {
    case MODEL_MATRIX:
    break;

    case MODEL_3DLUT:
    case MODEL_3DLUT_v2:
      if ( !cToneMappingFile )
        m_lutRef      = new TComLut3D( getNbp() );
      m_lutCoded      = new TComLut3D( getNbp(), true );
      m_lutResiduals  = new TComLut3D( getNbp() );
    break;
  }
}

TComToneMapping::~TComToneMapping()
{
	delete m_pcTempPic;
	if ( m_lutRef )       delete m_lutRef;
	if ( m_lutCoded )     delete m_lutCoded;
	if ( m_lutResiduals ) delete m_lutResiduals;

  if ( m_gain ) {
    for ( Int i=0 ; i<3 ; i++ ) free( m_gain[i] );

    free( m_gain );
  }
  if ( m_offset ) free( m_offset );
}

Void  TComToneMapping::writeFile( Char *cFile, TComLut3D* lut )
{
  Char  cTmp[500];

  TComLut3D*  plut = lut ? lut : m_lutRef ;

  if ( m_model == MODEL_MATRIX )
    sprintf( cTmp, "%s_Matrix.txt", cFile );
  else
    sprintf( cTmp, "%s_3Dlut_nbp%d.txt", cFile, plut->getNbp() );

  FILE  *fic = fopen( cTmp, "w" );

  if ( m_model == MODEL_MATRIX )            fprintf( fic, "MATRIX\n");
  if ( m_model == MODEL_3DLUT )             fprintf( fic, "3DLUT\n");
  if ( m_model == MODEL_3DLUT_v2 )          fprintf( fic, "3DLUT_v2\n");

  // Matrix / Gain model
  if ( m_model == MODEL_MATRIX ) {
    fprintf(fic, "%d\n", m_denom);

    for ( Int i=0 ; i<3 ; i++ ) {
      for ( Int j=0 ; j<3 ; j++ ) 
        fprintf(fic, "%d\t", m_gain[i][j]);
      fprintf(fic, "%d\n", m_offset[i]);
    }
  }
  // 3D LUT model
  else {
    fprintf(fic, "%d\n", m_nbpCode);
    for ( Int r=0 ; r<plut->getNbp() ; r++ )
    for ( Int g=0 ; g<plut->getNbp() ; g++ ) {
      for ( Int b=0 ; b<plut->getNbp() ; b++ ) {
        node* n = plut->getNode( r,g,b );
        fprintf(fic, "%d ",  n->sY);
        fprintf(fic, "%d ",  n->sU);
        fprintf(fic, "%d\t\t", n->sV);
      }
      fprintf(fic, "\n");
    }
  }

  fclose(fic);
}

Void  TComToneMapping::loadFile( Char *cFile )
{
  FILE  *fic = fopen( cFile, "r" );

  Char  format[100];
  fscanf( fic, "%s", format );

  // Matrix / Gain model
  if ( strcmp(format,"MATRIX")==0 ) {
    m_model = MODEL_MATRIX;
    fscanf(fic, "%d", &m_denom);

    Bool  bAllocateMatrix = (!m_gain || !m_offset) ? (true) : (false);
    if ( bAllocateMatrix ) {
      m_gain    = (Int **)malloc(sizeof(Int *) * 3 );
      m_offset  = (Int  *)malloc(sizeof(Int)   * 3 );
    }

    for ( Int i=0 ; i<3 ; i++ ) {
      if ( bAllocateMatrix )
        m_gain[i]  = (Int *)malloc(sizeof(Int) * 3 );

      for ( Int j=0 ; j<3 ; j++ ) 
        fscanf(fic, "%d", &m_gain[i][j]);
      fscanf(fic, "%d", &m_offset[i]);
    }
  }
  // 3D LUT model
  else if ( strcmp(format,"3DLUT")==0 || strcmp(format,"3DLUT_v2")==0 ) {
    if (  strcmp(format,"3DLUT")==0 ) {
      m_model = MODEL_3DLUT;
      fscanf(fic, "%d", &m_nbpCode);
    }
    else {
      m_model = MODEL_3DLUT_v2;
      fscanf(fic, "%d", &m_nbpCode );
      fscanf(fic, "%d", &m_denom_shift );
    }
    
    m_lutRef  = new TComLut3D( getNbp() );

    for ( Int r=0 ; r<getNbp() ; r++ )
    for ( Int g=0 ; g<getNbp() ; g++ )
    for ( Int b=0 ; b<getNbp() ; b++ ) {
      node* n = m_lutRef->getNode( r,g,b );
      fscanf(fic, "%hi", &n->sY);
      fscanf(fic, "%hi", &n->sU);
      fscanf(fic, "%hi", &n->sV);
    }
  }

  fclose(fic);
}

Void  TComToneMapping::modelMatrix( Int** matrix, Int* offset, Int denom, node *nodeIn, node *nodeOut )
{
  Int rOff  = denom >> 1;

  nodeOut->sY = offset[0] + ( matrix[0][0]*nodeIn->sY + matrix[0][1]*nodeIn->sU + matrix[0][2]*nodeIn->sV + rOff ) / denom;
  nodeOut->sU = offset[1] + ( matrix[1][0]*nodeIn->sY + matrix[1][1]*nodeIn->sU + matrix[1][2]*nodeIn->sV + rOff ) / denom;
  nodeOut->sV = offset[2] + ( matrix[2][0]*nodeIn->sY + matrix[2][1]*nodeIn->sU + matrix[2][2]*nodeIn->sV + rOff ) / denom;
}

Void	TComToneMapping::toneMap( TComPicYuv* pcBasePic, TComLut3D *lut, Bool bMarkUsedVertices )
{
  printf("toneMap (%s model)...\n", (m_model==MODEL_MATRIX) ? "MATRIX" : "3D-LUT" );

  if ( !lut ) lut = m_lutCoded;

  if ( lut )  lut->setBitDepth( m_inputBitDepth, m_outputBitDepth, m_lutBitDepth );

	if ( !m_pcTempPic ) {
	  m_pcTempPic = new TComPicYuv;
    m_pcTempPic->create( pcBasePic->getWidth(), pcBasePic->getHeight() );
	}
  else
    assert( "TComToneMapping::toneMap() assert failed.", pcBasePic->getWidth()==m_pcTempPic->getWidth() && pcBasePic->getHeight()==m_pcTempPic->getHeight() );


  Int iHeight   = pcBasePic->getHeight();
  Int iWidth    = pcBasePic->getWidth();
  Int iStride   = pcBasePic->getStride();
  Int iCStride  = pcBasePic->getCStride();

  Pel*	Lum0    = pcBasePic->getLumaAddr();
  Pel*	Cb0	    = pcBasePic->getCbAddr();
  Pel*	Cr0	    = pcBasePic->getCrAddr();
  Pel*	Lum1    = m_pcTempPic->getLumaAddr();
  Pel*	Cb1	    = m_pcTempPic->getCbAddr();
  Pel*	Cr1	    = m_pcTempPic->getCrAddr();

  //Int offset  = m_denom_shift ? (1 << (m_denom_shift-1)) : (0);
/*
  node  nodeIn, nodeOut;
  for( Int y = 0; y < iHeight ; y++ ) {
    for( Int x = 0; x < iWidth ; x++ ) {
      nodeIn.sY = Lum0[x];
      nodeIn.sU = Cb0[x>>1];
      nodeIn.sV = Cr0[x>>1];

      if ( m_model == MODEL_MATRIX )  modelMatrix( m_gain, m_offset, m_denom, &nodeIn, &nodeOut );
#if TETRAHEDRAL
      else                            lut->interpol_tetra_ibd( &nodeIn, &nodeOut, inputBitDepth );
#else
      //else                            lut->interpol( &nodeIn, &nodeOut );
      else                            lut->interpol_ibd( &nodeIn, &nodeOut, inputBitDepth );
#endif

      //nodeOut.sY  = (nodeOut.sY + offset) >> m_denom_shift;
      //nodeOut.sU  = (nodeOut.sU + offset) >> m_denom_shift;
      //nodeOut.sV  = (nodeOut.sV + offset) >> m_denom_shift;

      Lum1[x]   = xClip( nodeOut.sY, inputBitDepth );
      Cb1[x>>1] = xClip( nodeOut.sU, inputBitDepth );
      Cr1[x>>1] = xClip( nodeOut.sV, inputBitDepth );

      //if ( y==(iHeight-1) && x==0 ) printf("DEBUG\tTComToneMapping::toneMap():\t%d %d %d\n", Lum1[x], Cb1[x>>1], Cr1[x>>1] );
    }
    Lum0  += iStride;
    Lum1  += iStride;
    if ( (y % 2) ) {
      Cb0 += iCStride;
      Cr0 += iCStride;
      Cb1 += iCStride;
      Cr1 += iCStride;
    }
  }
*/
  Pel*  LumPrev0;
  node  nodeIn, nodeOut;
  for( Int y = 0; y < iHeight ; y++ ) {
    Bool  bDoChroma = (y % 2);
    for( Int x = 0; x < iWidth ; x++ ) {
      nodeIn.sY = Lum0[x];
      nodeIn.sU = Cb0[x>>1];
      nodeIn.sV = Cr0[x>>1];

      if ( bDoChroma && (x%2) ) {
      //if ( true ) {
        if ( m_model == MODEL_MATRIX )  modelMatrix( m_gain, m_offset, m_denom, &nodeIn, &nodeOut );
#if TETRAHEDRAL
        else {
          //lut->interpol_tetra_ibd( &nodeIn, &nodeOut );

          lut->interpol_tetra_Lum( &nodeIn, &nodeOut, bMarkUsedVertices );
          Short sY = (Lum0[x] + Lum0[x-1] + LumPrev0[x] + LumPrev0[x-1] + 2) >> 2 ;
          lut->interpol_tetra_Chroma( sY, &nodeIn, &nodeOut );
        }
#else
      //else                            lut->interpol( &nodeIn, &nodeOut );
      else                            lut->interpol_ibd( &nodeIn, &nodeOut );
#endif

        Lum1[x]   = xClip( nodeOut.sY, m_outputBitDepth );
        Cb1[x>>1] = xClip( nodeOut.sU, m_outputBitDepth );
        Cr1[x>>1] = xClip( nodeOut.sV, m_outputBitDepth );
      }
      else {
        if ( m_model == MODEL_MATRIX )  modelMatrix( m_gain, m_offset, m_denom, &nodeIn, &nodeOut );
#if TETRAHEDRAL
        else                            lut->interpol_tetra_Lum( &nodeIn, &nodeOut, bMarkUsedVertices );
#else
        else                            lut->interpol_ibd( &nodeIn, &nodeOut );
#endif

        Lum1[x]   = xClip( nodeOut.sY, m_outputBitDepth );
      }

    }

    LumPrev0 = Lum0;
    Lum0  += iStride;
    Lum1  += iStride;
    if ( (y % 2) ) {
      Cb0 += iCStride;
      Cr0 += iCStride;
      Cb1 += iCStride;
      Cr1 += iCStride;
    }
  }


  //dumpPicture("toneMapped", m_pcTempPic, 0 );

  printf("toneMap done.\n");
}

Void  TComToneMapping::dumpPicture(Char* fileName, TComPicYuv*	pPic, Int iShift )
{
  Char  name[1000];
  sprintf(name, "%s_%dx%d.yuv", fileName, pPic->getWidth(), pPic->getHeight());
  FILE	*fic = fopen(name, "wb");
  Pel *p = pPic->getLumaAddr();
  Int iStride = pPic->getStride();
  UChar*  uBuf = new UChar [pPic->getWidth()];
  for ( Int i=0 ; i<pPic->getHeight() ; i++ ) {
    for ( Int j=0 ; j<pPic->getWidth() ; j++ )
      uBuf[j] = (UChar)( p[j] >> iShift );
    fwrite( uBuf, sizeof(UChar), pPic->getWidth(), fic);
    p += iStride;
  }

  p = pPic->getCbAddr();
  iStride = pPic->getCStride();
  for ( Int i=0 ; i<pPic->getHeight()>>1 ; i++ ) {
    for ( Int j=0 ; j<pPic->getWidth()>>1 ; j++ )
      uBuf[j] = (UChar)( p[j] >> iShift );
    fwrite( uBuf, sizeof(UChar), pPic->getWidth()>>1, fic);
    p += iStride;
  }

  p = pPic->getCrAddr();
  iStride = pPic->getCStride();
  for ( Int i=0 ; i<pPic->getHeight()>>1 ; i++ ) {
    for ( Int j=0 ; j<pPic->getWidth()>>1 ; j++ )
      uBuf[j] = (UChar)( p[j] >> iShift );
    fwrite( uBuf, sizeof(UChar), pPic->getWidth()>>1, fic);
    p += iStride;
  }

  delete uBuf;

  fclose(fic);
}

inline  Int TComToneMapping::getNbp()
{
  return nbpCode[m_nbpCode];
}

Void	TComToneMapping::decode()
{
  m_lutCoded->reset();
  m_lutCoded->decode( 0, 0,0,0, m_errQ, m_lutResiduals, m_lutCoded->getOctTree(), m_bVerbose );
}
